﻿Option Strict On

Imports Nist.Logging
Imports System.IO
Imports System.Xml
Imports System.Text
Imports System.ServiceModel.Web
Imports System.ServiceModel.ServiceBehaviorAttribute

<ServiceModel.ServiceBehavior(IncludeExceptionDetailInFaults:=True, Namespace:="http://docs.oasis-open.org/bias/ns/bias-1.0/")>
Public Class BIASService
    Implements BIAS_v1, IPolicyRetriever

    Public Sub New()
    End Sub

    Public Function QueryCapabilities(ByVal QueryCapabilitiesRequest As QueryCapabilitiesRequest) As QueryCapabilitiesResponsePackage Implements BIAS_v1.QueryCapabilities
        Debug("QueryCapabilities()")
        Dim Response As New QueryCapabilitiesResponsePackage()

        Response.ResponseStatus = New ResponseStatus
        Response.ResponseStatus.Return = 0

        Response.CapabilityList = New CapabilityListType()
        Dim cap As New CapabilityType()
        cap.CapabilityName = CapabilityName.AggregateInputDataOptional
        Response.CapabilityList.Add(cap)
        cap = New CapabilityType()
        cap.CapabilityName = CapabilityName.AggregateInputDataRequired
        Response.CapabilityList.Add(cap)
        cap = New CapabilityType()
        cap.CapabilityName = CapabilityName.AggregateProcessingOption
        Response.CapabilityList.Add(cap)
        cap = New CapabilityType()
        cap.CapabilityName = CapabilityName.AggregateReturnData
        Response.CapabilityList.Add(cap)
        cap = New CapabilityType()
        cap.CapabilityName = CapabilityName.AggregateServiceDescription
        Response.CapabilityList.Add(cap)
        cap = New CapabilityType()
        cap.CapabilityName = CapabilityName.BiographicDataSet
        Response.CapabilityList.Add(cap)
        cap = New CapabilityType()
        cap.CapabilityName = CapabilityName.CBEFFPatronFormat
        Response.CapabilityList.Add(cap)
        cap = New CapabilityType()
        cap.CapabilityName = CapabilityName.ConformanceClass
        Response.CapabilityList.Add(cap)
        cap = New CapabilityType()
        cap.CapabilityName = CapabilityName.IdentityModel
        Response.CapabilityList.Add(cap)
        cap = New CapabilityType()
        cap.CapabilityName = CapabilityName.SupportedBiometric
        Response.CapabilityList.Add(cap)

        Return Response
    End Function

    Public Function Enroll(ByVal EnrollRequest As EnrollRequest) As EnrollResponsePackage Implements BIAS_v1.Enroll
        Debug("Enroll()")
        Info("Processing new enrollment")

        Dim Result As New EnrollResponsePackage()
        Result.ResponseStatus = New ResponseStatus

        If Gallery.Subjects.ContainsKey(EnrollRequest.InputData.GUID) Then
            Result.ResponseStatus.Message = "Subject with the specified identification key is already enrolled in this system."
            Info(Result.ResponseStatus.Message)
            Result.ResponseStatus.Return = 1
        Else
            Dim Enrollee As New Subject()

            Enrollee.GUID = EnrollRequest.InputData.GUID
            Enrollee.GivenName = EnrollRequest.InputData.GivenName
            Enrollee.FamilyName = EnrollRequest.InputData.FamilyName
            Enrollee.DateOfBirth = EnrollRequest.InputData.DateOfBirth
            Enrollee.Sex = EnrollRequest.InputData.Sex
            Enrollee.Citizenship = EnrollRequest.InputData.Citizenship

            If EnrollRequest.InputData.Images IsNot Nothing Then
                For Each CurrentSample As Image In EnrollRequest.InputData.Images
                    Dim ContentType As String = CurrentSample.ContentType
                    Dim FingerPosition As String = CurrentSample.FingerPosition
                    Dim HandDesignation As String = CurrentSample.HandPosition
                    Dim ImageData As Byte() = CurrentSample.ImageData

                    If ImageData IsNot Nothing AndAlso ImageData.Length > 0 Then
                        Dim Sample As Drawing.Bitmap = ProcessImage(ImageData, ContentType)

                        If Sample Is Nothing Then
                            Result.ResponseStatus.Return = 2
                            Result.ResponseStatus.Message = String.Format("Specified content type '{0}' unsupported.", ContentType)
                            Info(Result.ResponseStatus.Message)

                            Return Result
                        End If

                        Dim Minutiae As Nbis.Minutia() = Nbis.DetectMinutiae.FromBitmap(Sample, 500)

                        If Enrollee.MinutiaeTemplates Is Nothing Then
                            Enrollee.MinutiaeTemplates = New ArrayList()
                        End If

                        Enrollee.MinutiaeTemplates.Add(New FingerprintTemplate(HandDesignation, FingerPosition, Minutiae))
                    End If
                Next

                Gallery.UpdateSubject(Enrollee.GUID, Enrollee)
                Result.ResponseStatus.Return = 0
                Result.ResponseStatus.Message = "Enrollment successful"
                Info(Result.ResponseStatus.Message)
            End If

        End If

        Return Result
    End Function

    Public Function ProcessImage(ByRef ImageData As Byte(), ByVal ContentType As String) As Drawing.Bitmap
        Dim Image As Drawing.Bitmap = Nothing

        If (ContentType.Equals("image/wsq", StringComparison.CurrentCultureIgnoreCase)) Then
            Image = Nbis.Wsq.FromWsqToBitmap(ImageData)

        ElseIf (ContentType.Equals("image/bmp", StringComparison.CurrentCultureIgnoreCase)) Then
            Dim Memory As New System.IO.MemoryStream(ImageData)
            Image = CType(Drawing.Image.FromStream(Memory), Drawing.Bitmap)
        Else

        End If

        Return Image
    End Function

    Public Function Verify(ByVal VerifyRequest As VerifyRequest) As VerifyResponsePackage Implements BIAS_v1.Verify
        Debug("Verify()")
        Info("Processing verification")

        Dim IdentityClaim As String = VerifyRequest.Identity.IdentityClaim
        Debug(" - Identity Claim = {0}", IdentityClaim)

        Dim Result As New VerifyResponsePackage()
        Result.ResponseStatus = New ResponseStatus
        Result.ResponseStatus.Return = 0
        Result.ResponseStatus.Message = ""

        If Gallery.Subjects.ContainsKey(IdentityClaim) Then
            Dim Target As Subject = Gallery.Subjects(IdentityClaim)
            Dim TargetMinutiae As Nbis.Minutia() = Nothing

            If (VerifyRequest.InputData.Images Is Nothing OrElse VerifyRequest.InputData.Images.Count = 0) Then
                Result.ResponseStatus.Message = "Verification failed; no probe sample provided"
                Result.ResponseStatus.Return = 2
            Else
                Dim FingerPosition As String = VerifyRequest.InputData.Images(0).FingerPosition
                Dim HandDesignation As String = VerifyRequest.InputData.Images(0).HandPosition

                For Each MinutiaeTemplate As FingerprintTemplate In Target.MinutiaeTemplates
                    If MinutiaeTemplate.FingerPosition.Equals(FingerPosition, StringComparison.CurrentCultureIgnoreCase) AndAlso MinutiaeTemplate.HandDesignation.Equals(HandDesignation, StringComparison.CurrentCultureIgnoreCase) Then
                        TargetMinutiae = MinutiaeTemplate.Minutiae
                        Exit For
                    End If
                Next MinutiaeTemplate

                If TargetMinutiae IsNot Nothing Then
                    Dim ProbeImage As Drawing.Bitmap = ProcessImage(VerifyRequest.InputData.Images(0).ImageData, VerifyRequest.InputData.Images(0).ContentType)
                    Dim ProbeMinutiae As Nbis.Minutia() = Nbis.DetectMinutiae.FromBitmap(ProbeImage, 500)

                    Dim Match As Integer = Nbis.Matcher.Match(ProbeMinutiae, TargetMinutiae)
                    Result.Score = Nbis.Matcher.MatchScore(ProbeMinutiae, TargetMinutiae)
                    Result.Match = CBool(IIf(Result.Score >= MatchScoreThreshold, True, False))

                    If Result.Match Then
                        Result.ResponseStatus.Message = String.Format("Verification successful; subject and target match. Match={0};MatchScore={1};", Match, Result.Score)
                        Result.ResponseStatus.Return = 0
                    Else
                        Result.ResponseStatus.Message = "Verification failed; subject and target do not match"
                        Result.ResponseStatus.Return = 3
                    End If
                Else
                    Result.Match = False
                    Result.Score = -1
                    Result.ResponseStatus.Return = 4
                    Result.ResponseStatus.Message = "Verification failed; specified subject does not have corresponding biometric data"
                End If
            End If
        Else
            Result.ResponseStatus.Message = "Verification failed; claimed identity is not enrolled"
            Result.ResponseStatus.Return = 1
        End If

        Info(Result.ResponseStatus.Message)
        Return Result
    End Function

    Public Function RetrieveInformation(ByVal RetrieveInformationRequest As RetrieveInformationRequest) As RetrieveInformationResponsePackage Implements BIAS_v1.RetrieveInformation
        Debug("RetrieveInformation()")
        Info("Retrieving information")

        Dim Result As New RetrieveInformationResponsePackage()
        Result.ResponseStatus = New ResponseStatus()

        Dim SubjectID As String = RetrieveInformationRequest.Identity.SubjectID

        If Gallery.Subjects.ContainsKey(SubjectID) Then
            Dim Information As Subject = Gallery.Subjects(SubjectID)

            Result.ResponseStatus.Message = "Target is enrolled; returning information"
            Result.ResponseStatus.Return = 0

            Result.ReturnData = New InformationType()
            Result.ReturnData.GivenName = Information.GivenName
            Result.ReturnData.FamilyName = Information.FamilyName
            Result.ReturnData.GUID = Information.GUID
            Result.ReturnData.Sex = Information.Sex
            Result.ReturnData.Citizenship = Information.Citizenship
            Result.ReturnData.DateOfBirth = Information.DateOfBirth

        Else
            Result.ResponseStatus.Message = "Target is not enrolled; no information will be returned"
            Result.ResponseStatus.Return = 1
        End If

        Info(Result.ResponseStatus.Message)
        Return Result
    End Function

    Public Function Identify(ByVal IdentifyRequest As IdentifyRequest) As IdentifyResponsePackage Implements BIAS_v1.Identify
        Debug("Identify()")
        Info("Identifying subject")

        Dim Result As New IdentifyResponsePackage()
        Result.ResponseStatus = New ResponseStatus()
        Result.ResponseStatus.Message = Nothing
        Result.ResponseStatus.Return = 0

        If (IdentifyRequest Is Nothing) Then
            Result.ResponseStatus.Message = "There is no request data found for this operation"
            Result.ResponseStatus.Return = 1
        End If

        If (IdentifyRequest.InputData Is Nothing) Then
            Result.ResponseStatus.Message = "The input data for this request cannot be found"
            Result.ResponseStatus.Return = 2
        End If

        If (IdentifyRequest.InputData.Images Is Nothing OrElse IdentifyRequest.InputData.Images.Count = 0) Then
            Result.ResponseStatus.Message = "There are no biometric samples to identify"
            Result.ResponseStatus.Return = 3
        End If

        If (Result.ResponseStatus.Message IsNot Nothing) Then
            Info(Result.ResponseStatus.Message)
            Return Result
        End If

        Dim ContentType As String = IdentifyRequest.InputData.Images(0).ContentType
        Dim ImageData As Byte() = IdentifyRequest.InputData.Images(0).ImageData

        Dim FingerPosition As String = IdentifyRequest.InputData.Images(0).FingerPosition
        Dim HandDesignation As String = IdentifyRequest.InputData.Images(0).HandPosition

        Dim ProbeImage As Drawing.Bitmap = ProcessImage(ImageData, ContentType)
        Dim ProbeMinutiae As Nbis.Minutia() = Nbis.DetectMinutiae.FromBitmap(ProbeImage, 500)

        Dim Identifications As New List(Of Subject)

        For Each TargetSubject In Gallery.Subjects
            Debug("new target")
            For Each TargetTemplate As FingerprintTemplate In TargetSubject.Value.MinutiaeTemplates
                Dim PerformComparison As Boolean = False

                If (FingerPosition Is Nothing AndAlso HandDesignation Is Nothing) Then
                    'compare probe to all fingers
                    PerformComparison = False
                ElseIf (TargetTemplate.FingerPosition.Equals(FingerPosition, StringComparison.CurrentCultureIgnoreCase) AndAlso TargetTemplate.HandDesignation.Equals(HandDesignation, StringComparison.CurrentCultureIgnoreCase)) Then
                    'compare probe with this target
                    PerformComparison = True
                ElseIf TargetTemplate.FingerPosition.Equals(FingerPosition, StringComparison.CurrentCultureIgnoreCase) Then
                    'compare probe with X finger on both hands
                    PerformComparison = False
                ElseIf TargetTemplate.HandDesignation.Equals(HandDesignation, StringComparison.CurrentCultureIgnoreCase) Then
                    'compare probe with all fingers on X hand
                    PerformComparison = False
                End If

                If PerformComparison Then
                    Dim Match As Integer = Nbis.Matcher.Match(ProbeMinutiae, TargetTemplate.Minutiae)
                    Dim MatchScore As Integer = Nbis.Matcher.MatchScore(ProbeMinutiae, TargetTemplate.Minutiae)
                    Debug("{0} = {1}", Match, MatchScore)
                    If MatchScore >= MatchScoreThreshold Then
                        Info("Possible identification as {0} [{1}]", TargetSubject.Value.Name, TargetSubject.Key)
                        Identifications.Add(TargetSubject.Value)
                        Exit For
                    End If
                End If
            Next TargetTemplate
        Next TargetSubject

        If Identifications.Count > 0 Then
            Result.ResponseStatus.Return = 0
            Result.ResponseStatus.Message = String.Format("Subject identified as {0} possible identification(s)", Identifications.Count)

            Result.ReturnData = New InformationType()
            Result.ReturnData.Identities = New InformationType.IdentitiesType
            For Each PossibleIdentification As Subject In Identifications
                Result.ReturnData.Identities.Add(PossibleIdentification.GUID)
            Next
        Else
            Result.ResponseStatus.Return = 4
            Result.ResponseStatus.Message = "Subject could not be identified"
        End If

        Info(Result.ResponseStatus.Message)
        Return Result
    End Function






    'Asynchronous
    Public Function GetEnrollResults(ByVal GetEnrollResultsRequest As GetEnrollResultsRequest) As GetEnrollResultsResponsePackage Implements BIAS_v1.GetEnrollResults
        Debug("GetEnrollResults()")
        Return Nothing
    End Function

    'Asynchronous
    Public Function GetIdentifyResults(ByVal GetIdentifyResultsRequest As GetIdentifyResultsRequest) As GetIdentifyResultsResponsePackage Implements BIAS_v1.GetIdentifyResults
        Debug("GetIdentifyResults()")
        Return Nothing
    End Function

    'Asynchronous
    Public Function GetVerifyResults(ByVal GetVerifyResultsRequest As GetVerifyResultsRequest) As GetVerifyResultsResponsePackage Implements BIAS_v1.GetVerifyResults
        Debug("GetVerifyResults()")
        Return Nothing
    End Function


    ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' '
    ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' '

    Public Function AddSubjectToGallery(ByVal AddSubjectToGalleryRequest As AddSubjectToGalleryRequest) As AddSubjectToGalleryResponsePackage Implements BIAS_v1.AddSubjectToGallery
        Debug("AddSubjectToGallery()")
        Return Nothing
    End Function

    Public Function CheckQuality(ByVal CheckQualityRequest As CheckQualityRequest) As CheckQualityResponsePackage Implements BIAS_v1.CheckQuality
        Debug("CheckQuality()")
        Return Nothing
    End Function

    Public Function ClassifyBiometricData(ByVal ClassifyBiometricDataRequest As ClassifyBiometricDataRequest) As ClassifyBiometricDataResponsePackage Implements BIAS_v1.ClassifyBiometricData
        Debug("ClassifyBiometricData()")
        Return Nothing
    End Function

    Public Function CreateSubject(ByVal CreateSubjectRequest As CreateSubjectRequest) As CreateSubjectResponsePackage Implements BIAS_v1.CreateSubject
        Debug("CreateSubject()")
        Return Nothing
    End Function

    Public Function DeleteBiographicData(ByVal DeleteBiographicDataRequest As DeleteBiographicDataRequest) As DeleteBiographicDataResponsePackage Implements BIAS_v1.DeleteBiographicData
        Debug("DeleteBiographicData()")
        Return Nothing
    End Function

    Public Function DeleteBiometricData(ByVal DeleteBiometricDataRequest As DeleteBiometricDataRequest) As DeleteBiometricDataResponsePackage Implements BIAS_v1.DeleteBiometricData
        Debug("DeleteBiometricData()")
        Return Nothing
    End Function

    Public Function DeleteSubject(ByVal DeleteSubjectRequest As DeleteSubjectRequest) As DeleteSubjectResponsePackage Implements BIAS_v1.DeleteSubject
        Debug("DeleteSubject()")
        Return Nothing
    End Function

    Public Function DeleteSubjectFromGallery(ByVal DeleteSubjectFromGalleryRequest As DeleteSubjectFromGalleryRequest) As DeleteSubjectFromGalleryResponsePackage Implements BIAS_v1.DeleteSubjectFromGallery
        Debug("DeleteSubjectFromGallery()")
        Return Nothing
    End Function

    Public Function GetIdentifySubjectResults(ByVal GetIdentifySubjectResultsRequest As GetIdentifySubjectResultsRequest) As GetIdentifySubjectResultsResponsePackage Implements BIAS_v1.GetIdentifySubjectResults
        Debug("GetIdentifySubjectResults()")
        Return Nothing
    End Function

    Public Function IdentifySubject(ByVal IdentifySubjectRequest As IdentifySubjectRequest) As IdentifySubjectResponsePackage Implements BIAS_v1.IdentifySubject
        Debug("IdentifySubject()")
        Return Nothing
    End Function

    Public Function ListBiographicData(ByVal ListBiographicDataRequest As ListBiographicDataRequest) As ListBiographicDataResponsePackage Implements BIAS_v1.ListBiographicData
        Debug("ListBiographicData()")
        Return Nothing
    End Function

    Public Function ListBiometricData(ByVal ListBiometricDataRequest As ListBiometricDataRequest) As ListBiometricDataResponsePackage Implements BIAS_v1.ListBiometricData
        Debug("ListBiometricData()")
        Return Nothing
    End Function

    Public Function PerformFusion(ByVal PerformFusionRequest As PerformFusionRequest) As PerformFusionResponsePackage Implements BIAS_v1.PerformFusion
        Debug("PerformFusion()")
        Return Nothing
    End Function

    Public Function RetrieveBiographicInformation(ByVal RetrieveBiographicInformationRequest As RetrieveBiographicInformationRequest) As RetrieveBiographicInformationResponsePackage Implements BIAS_v1.RetrieveBiographicInformation
        Debug("RetrieveBiographicInformation()")
        Return Nothing
    End Function

    Public Function RetrieveBiometricInformation(ByVal RetrieveBiometricInformationRequest As RetrieveBiometricInformationRequest) As RetrieveBiometricInformationResponsePackage Implements BIAS_v1.RetrieveBiometricInformation
        Debug("RetrieveBiometricInformation()")
        Return Nothing
    End Function

    Public Function SetBiographicData(ByVal SetBiographicDataRequest As SetBiographicDataRequest) As SetBiographicDataResponsePackage Implements BIAS_v1.SetBiographicData
        Debug("SetBiographicData()")
        Return Nothing
    End Function

    Public Function SetBiometricData(ByVal SetBiometricDataRequest As SetBiometricDataRequest) As SetBiometricDataResponsePackage Implements BIAS_v1.SetBiometricData
        Debug("SetBiometricData()")
        Return Nothing
    End Function

    Public Function TransformBiometricData(ByVal TransformBiometricDataRequest As TransformBiometricDataRequest) As TransformBiometricDataResponsePackage Implements BIAS_v1.TransformBiometricData
        Debug("TransformBiometricData()")
        Return Nothing
    End Function

    Public Function UpdateBiographicData(ByVal UpdateBiographicDataRequest As UpdateBiographicDataRequest) As UpdateBiographicDataResponsePackage Implements BIAS_v1.UpdateBiographicData
        Debug("UpdateBiographicData()")
        Return Nothing
    End Function

    Public Function UpdateBiometricData(ByVal UpdateBiometricDataRequest As UpdateBiometricDataRequest) As UpdateBiometricDataResponsePackage Implements BIAS_v1.UpdateBiometricData
        Debug("UpdateBiometricData()")
        Return Nothing
    End Function

    Public Function VerifySubject(ByVal VerifySubjectRequest As VerifySubjectRequest) As VerifySubjectResponsePackage Implements BIAS_v1.VerifySubject
        Debug("VerifySubject()")
        Return Nothing
    End Function


    Public Function GetPolicy() As Stream Implements IPolicyRetriever.GetPolicy
        '<cross-domain-policy xsi:noNamespaceSchemaLocation="http://www.adobe.com/xml/schemas/PolicyFile.xsd">
        '<allow-access-from domain="twitter.com"/>
        '<allow-access-from domain="api.twitter.com"/>
        '<allow-access-from domain="search.twitter.com"/>
        '<allow-access-from domain="static.twitter.com"/>
        '<site-control permitted-cross-domain-policies="master-only"/>
        '<allow-http-request-headers-from domain="*.twitter.com" headers="*" secure="true"/>
        '</cross-domain-policy>

        Dim policy = "<cross-domain-policy>" &
                     "  <allow-access-from domain='*' />" &
                     "  <site-control permitted-cross-domain-policies='all' />" &
                     "  <allow-http-request-headers-from domain='*' headers='*' />" &
                     "</cross-domain-policy>"

        WebOperationContext.Current.OutgoingResponse.ContentType = "application/xml"
        Return New MemoryStream(Encoding.UTF8.GetBytes(policy))
    End Function

End Class